/*
 * ParametricSystem.java
 *
 * Created on August 2, 2006, 3:01 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm.data;

import sddm.*;
import java.util.*;

/**
 * This class is a representation of a parametric system - that is, one where a function applied over some range results in a set of points.
 * @author Erik J. Tollerud
 */
public class ParametricSystem {
    private PointList points;
    private ParametricGenerator gen;
    private Vector<Float> time;
    private int currIter;
    
    /** Creates a new instance of ParametricSystem using the default linear generator */
    public ParametricSystem() {
        this(new DefaultLinearGenerator());
    }
    
    /**
     * Creates a new instance of ParametricSystem using
     * a linear generator with the specified slopes: i.e. x = deltaX*i. Starts at {0,0,0}
     * @param deltaX the x-component slope
     * @param deltaY the y-component slope
     * @param deltaZ the z-component slope
     */
    public ParametricSystem(float deltaX, float deltaY, float deltaZ) {
        this(new DefaultLinearGenerator(deltaX, deltaY, deltaZ));
    }
    
    /**
     * Creates a new instance of ParametricSystem using
     * a linear generator with the specified slopes and time step size: i.e. x = deltaX*deltaT*i. Starts at {0,0,0}
     * @param deltaX the x-component slope
     * @param deltaY the y-component slope
     * @param deltaZ the z-component slope
     * @param deltaT the time step size per iteration
     */
    public ParametricSystem(float deltaX, float deltaY, float deltaZ, float deltaT) {
        this(new DefaultLinearGenerator(deltaX, deltaY, deltaZ,deltaT));
    }
    
    /**
     * Creates a new instance of ParametricSystem using
     * a linear generator with the specified slopes and initial conditions: i.e. x = deltaX*deltaT*i+x0.
     * @param deltaX the x-component slope
     * @param deltaY the y-component slope
     * @param deltaZ the z-component slope
     * @param x0 the x initial coordinate
     * @param y0 the y initial coordinate
     * @param z0 the z initial coordinate
     */
    public ParametricSystem(float deltaX, float deltaY, float deltaZ, float x0, float y0, float z0) {
        this(new DefaultLinearGenerator(deltaX,deltaY,deltaZ,x0,y0,z0));
    }
    
    /**
     * Creates a new instance of ParametricSystem using
     * a linear generator with the specified slopes, time step, and initial conditions: i.e. x = deltaX*(deltaT*i + t0)+x0.
     * @param deltaX the x-component slope
     * @param deltaY the y-component slope
     * @param deltaZ the z-component slope
     * @param deltaT the time step size per iteration 
     * @param x0 the x initial coordinate
     * @param y0 the y initial coordinate
     * @param z0 the z initial coordinate
     * @param t0 the starting time
     */
    public ParametricSystem(float deltaX, float deltaY, float deltaZ, float deltaT, float x0, float y0, float z0, float t0) {
        this(new DefaultLinearGenerator(deltaX,deltaY,deltaZ,deltaT,x0,y0,z0,t0));
    }
    
    /**
     * Creates a new instance of ParametricSystem using a specified generator
     * @param gen the generator to use for this parametric system
     * @throws SDDMException if the initial conditions of the supplied generator aren't all the same size
     */
    public ParametricSystem(ParametricGenerator gen) throws SDDMException {
        points = new PointList();
        time = new Vector<Float>();
        this.gen = gen;
        time.add(0.0f);
        addICs();
        currIter = 0;
    }
    
    private void addICs() throws SDDMException {
        float[] xs = gen.initialXs();
        float[] ys = gen.initialYs();
        float[] zs = gen.initialZs();
        float[] ws = gen.initialWs();
        float[] ts = gen.initialTs();
        if (xs.length != ys.length || 
                xs.length != zs.length || 
                xs.length != ws.length ||
                xs.length != ts.length ||
                ys.length != zs.length ||
                ys.length != ws.length ||
                ys.length != ts.length ||
                zs.length != ws.length ||
                zs.length != ts.length ||
                ws.length != ts.length)
            throw new SDDMException("Supplied ParametricGenerator  "+gen.getClass().getName()+" does not have matching numbers of initial conditions in each component");
        
        for(int i = 0; i < xs.length; ++i)  {
            points.addPoint(xs[i],ys[i],zs[i],ws[i]);
            time.add(ts[i]);
        }
    }
    
    public void iterate() {
        float[] results = gen.nextIteration(this.time, this.points);
        points.addPoint(results);
        currIter++;
    }
    
    public PointList getPointList() {
        return this.points;
    }
    
    public int getCurrentIteration() {
        return this.currIter;
    }
    
    /**
     * Iterates through a discrete number of steps in this parametric system
     * @param numOfIterations the number of iterations to perform
     * @throws sddm.SDDMException if NaN or an infinity is encountered in the iteration process
     */
    public void iterate(int numOfIterations) throws SDDMException {
        float[] results; //TODO: test if its better to put this up here or in the for loop
        for(int i = currIter; i < (numOfIterations + currIter); i++) {
            results = gen.nextIteration( this.time, this.points);
            for(float f : results) {
                if (f==Float.NaN||f==Float.NEGATIVE_INFINITY||f==Float.POSITIVE_INFINITY) {
                    this.currIter+=i;
                    float[] pointBeforeError = points.getLastPoint();
                    throw new SDDMException("Invalid number "+f+" resulted during iteration #"+this.currIter + " see console for lead-up to breakdown");
                }
            }
            points.addPoint(results);
        }
        currIter+=numOfIterations;
    }

    public void generateThrough(float tf) {
        while(time.lastElement() < tf) 
            this.iterate();
    }
    
    public void generateUntil(float tf) {
        while(time.lastElement() != tf) 
            this.iterate();
    }
    
    public void generateRange(float t0, float tf) throws SDDMException {
        generateRange(gen.initialXs()[0],gen.initialYs()[0],gen.initialZs()[0],gen.initialWs()[0],t0,tf);
    }
    
    public void generateRange(float x0, float y0, float z0, float t0, float tf) throws SDDMException {
        generateRange(x0,y0,z0,gen.initialWs()[0],t0,tf);
    }
    
    public void generateRange(float x0, float y0, float z0, float w0, float t0, float tf) throws SDDMException {
        if(gen.initialTs().length != 1)
            throw new SDDMException("Tried to generate a range with a generator that requires multiple ICs");
        
        points.clear();
        time.clear();
        this.currIter = 0;
        
        points.addPoint(x0,y0,z0,w0);
        time.add(t0);
        
        generateThrough(tf);
    }
    
    public void sortByTime() {
        sortByTime(true);
    }
    
    public void sortByTime(boolean ascending) {
        //TODO:implement
    }

    public void setGen(ParametricGenerator gen) {
        this.gen = gen;
    }

    public ParametricGenerator getGen() {
        return gen;
    }

    public int size() {
        return this.points.size();
    }
    
    public String toString() {
        String pStr = this.points.toString();
        
        return "Parametric System with generator "+gen.toString()+ "\nOn iteration #"+this.currIter+". Point list:\n"+pStr;
    }
}

class DefaultLinearGenerator implements ParametricGenerator {
    float deltaX;
    float deltaY;
    float deltaZ;
    float deltaT;
    float x0;
    float y0;
    float z0;
    float t0;
    
    DefaultLinearGenerator() {
        deltaX = 1.0f;
        deltaY = 1.0f;
        deltaZ = 1.0f;
        deltaT = 1.0f;
    }
    
    DefaultLinearGenerator(float deltaX, float deltaY, float deltaZ) {
        this.deltaX = deltaX;
        this.deltaY = deltaY;
        this.deltaZ = deltaZ;
        this.deltaY = 1.0f;
        
        this.x0 = 0.0f;
        this.y0 = 0.0f;
        this.z0 = 0.0f;
        this.t0 = 0.0f;
    }
    
    DefaultLinearGenerator(float deltaX, float deltaY, float deltaZ, float initialX, float initialY, float initialZ) {
        this.deltaX = deltaX;
        this.deltaY = deltaY;
        this.deltaZ = deltaZ;
        this.deltaT = 1.0f;
        
        this.x0 = initialX;
        this.y0 = initialY;
        this.z0 = initialZ;
        this.t0 = 0.0f;
    }
    
    DefaultLinearGenerator(float deltaX, float deltaY, float deltaZ, float deltaT) {
        this.deltaX = deltaX;
        this.deltaY = deltaY;
        this.deltaZ = deltaZ;
        this.deltaY = deltaT;
    }
    
    DefaultLinearGenerator(float deltaX, float deltaY, float deltaZ, float deltaT, float initialX, float initialY, float initialZ, float initialT) {
        this.deltaX = deltaX;
        this.deltaY = deltaY;
        this.deltaZ = deltaZ;
        this.deltaY = deltaT;
        
        this.x0 = initialX;
        this.y0 = initialY;
        this.z0 = initialZ;
        this.t0 = initialT;
    }

    public float[] nextIteration(Vector<Float> time, PointList wholeList) {
        float[] res = new float[4];
        float[] lastElem = wholeList.getLastPoint();
        res[0] = lastElem[0]+deltaX;
        res[1] = lastElem[1]+deltaY;
        res[2] = lastElem[2]+deltaZ;
        res[3] = 1.0f;
        time.add(time.lastElement()+deltaT);
        return res;
    }

    public float[] initialZs() {
        float[] res = new float[1];
        res[0] = z0;
        return res;
    }

    public float[] initialYs() {
        float[] res = new float[1];
        res[0] = y0;
        return res;
    }

    public float[] initialXs() {
        float[] res = new float[1];
        res[0] = x0;
        return res;
    }
    
    public float[] initialTs() {
        float[] res = new float[1];
        res[0] = t0;
        return res;
    }

    public float[] initialWs() {
        float[] res = {1.0f};
        return res;
    }
    
    
}