/*
 * PointMap.java
 *
 * Created on August 4, 2006, 6:36 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm.data;

import java.io.*;
import java.util.*;
import javax.vecmath.Vector3f;
import sddm.*;

/**
 * This class maps an ordered list of points to a generic value, allowing repeated points with different values.
 *
 * @author Erik J. Tollerud
 */
public class PointMap<T> extends PointList {
    private ArrayList<T> values;
    Comparator compor;
    //TODO:cache statistic values and invalidate them as appropriate
    //TODO: make thread-safe?
    
    /**
     * Creates an empty new instance of PointMap
     */
    public PointMap() {
        super();
        values = new ArrayList<T>();
        compor = null;
    }
    
    /**
     * Creates a new instance of PointMap with the specified set of points, and null values.
     * @param ps The points to include in this PointMap
     */
    public PointMap(PointSet ps) {
        super(ps);
        values = new ArrayList<T>();
        for (int i = 0; i < ps.size(); ++i)
            values.add(null);
    }
    
    /**
     * Creates a new instance of PointMap with the specified set of points, and the specified values.
     * @param ps The points to include in this PointMap
     * @param values the values to assign (must be the same length as points)
     * @throws sddm.SDDMException If the number of points doesn't match the number of values.
     */
    public PointMap(PointSet ps, T[] values) throws SDDMException {
        super(ps);
        if (ps.size() != values.length)
            throw new SDDMException("Values and points not same number of elements in making PointMap.");
        
        this.values = new ArrayList<T>();
        
        for (int i = 0; i < ps.size(); ++i)
            this.values.add(values[i]);
    }
    
    /**
     * Creates a new instance of PointMap with the specified set of points, and null values.
     * @param points The points to include in this PointMap as either 3-element or 4-element arrays.
     */
    public PointMap(float[][] points) {
        super(points);
        this.values = new ArrayList<T>();
        for (int i = 0; i < points.length; ++i)
            this.values.add(null);
    }
    
    /**
     *  Creates a new instance of PointMap with the specified set of points, and the specified values.
     * @param points the points to use (3 or 4-element float arrays)
     * @param values the values to assign (must be the same length as points)
     * @throws sddm.SDDMException If the number of points doesn't match the number of values.
     */
    public PointMap(float[][] points, T[] values) throws SDDMException {
        super(points);
        if (points.length != values.length)
            throw new SDDMException("Values and points not same number of elements in making PointMap.");
        
        this.values = new ArrayList<T>();
        
        for (int i = 0; i < points.length; ++i)
            this.values.add(values[i]);
    }
    
    /**
     * Sets the comparator to be used by this map's max/min methods
     * @param comp a comparator for the appropriate map type, or null to use the objects' compareTo method
     */
    public void setComparator(Comparator comp) {
        this.compor = comp;
    }
    
    /**
     * Retrieves the comparator used by this map's max/min methods
     * @return the comparator for this instance or null if the compareTo method is used
     */
    public Comparator getComparator() {
        return this.compor;
    }
    
    /**
     * Tests if the points and values of this PointMap match the supplied object.
     * @param obj the object to test.  If not PointMap, this method will return false.
     * @return true if the supplied PointMap has the same elements as this PointMap
     */
    public boolean equals(Object obj) {
        boolean superBool = super.equals(obj);
        
        if (obj instanceof PointMap)
            return superBool && values.equals(((PointMap)obj).values);
        else
            return false;
    }
    
    /**
     * Loads points from a list of points and gives them null values. See {@link PointList.loadPointsFromFile(File pointFile, boolean threeElement)}
     */
    public void loadPointsFromFile(File pointFile, boolean threeElement) throws FileFormatException, FileNotFoundException {
        super.loadPointsFromFile(pointFile, threeElement);
        for(int i = 0; i < list.size();++i)
            values.add(null);
    }
    /**
     * Loads points from a list of points and gives them null values. See {@link PointList.loadPointsFromFile(File pointFile)}
     */
    public void loadPointsFromFile(File pointFile) throws FileFormatException, FileNotFoundException {
        super.loadPointsFromFile(pointFile);
        for(int i = 0; i < list.size();++i)
            values.add(null);
    }
    
    /**
     * Sets the points and values in this PointMap, clearing any exisitng values.
     * @param points the points to use in this PointMap.
     * @param values the values to use in thie PointMap - must have the same number as the number of points.
     * @throws sddm.SDDMException if the number of points don't match or one of the supplied points is not 3 or 4 elements.
     */
    public void set(float[][] points, T[] values) throws SDDMException{
        if (values.length != points.length)
            throw new SDDMException("Number of points does not match number of values while attempting to set points in PointMap");
        
        try {
            super.setPoints(points);
            this.values.clear();
            for(int i = 0; i < points.length;++i)
                this.values.add(values[i]);
        } catch (Exception e) {
            this.list.clear();
            this.values.clear();
            throw new SDDMException("Problem setting points in PointMap - map will be cleared:"+e.getMessage(),e);
        }
    }
    
    /**
     * Sets the point and value at the specified index
     * @param i the index to use for setting the point and value.
     * @param point the point to use for the point at index i in the list in either {x,y,z} or {x,y,z,w} format.
     * @param value the value to set at index i.
     * @throws java.lang.IndexOutOfBoundsException if the specified index does not exist in this PointMap.
     * @throws sddm.SDDMException if the point is not a 3 or 4-element array
     */
    public void set( int i, float[] point, T value) throws SDDMException, IndexOutOfBoundsException {
        super.setPoint(i,point);
        this.values.set(i,value);
    }
    
    /**
     * Sets the point and value at the specified index. Leaves the w-coordinate unchanged.
     * @param i the index to use for setting the point and value.
     * @param x the x-coordinate to set at the specified index.
     * @param y the y-coordinate to set at the specified index.
     * @param z the z-coordinate to set at the specified index.
     * @param value the value to set at index i.
     * @throws java.lang.IndexOutOfBoundsException if the specified index does not exist in this PointMap.
     */
    public void set( int i, float x, float y, float z, T value) throws IndexOutOfBoundsException {
        super.setPoint(i,x,y,z);
        this.values.set(i,value);
    }
    
    /**
     * Sets the point and value at the specified index.
     * @param i the index to use for setting the point and value.
     * @param x the x-coordinate to set at the specified index.
     * @param y the y-coordinate to set at the specified index.
     * @param z the z-coordinate to set at the specified index.
     * @param w the w-coordinate to set at the specified index.
     * @param value the value to set at index i.
     * @throws java.lang.IndexOutOfBoundsException if the specified index does not exist in this PointMap.
     */
    public void set( int i, float x, float y, float z, float w, T value) throws IndexOutOfBoundsException {
        super.setPoint(i,x,y,z,w);
        this.values.set(i,value);
    }
    
    /**
     * Sets the points in this PointMap, clearing the exisitng points and setting all the values to null.
     * @param points The points to use as an array of 3 or 4-element arrays.
     */
    public void setPoints(float[][] points) {
        super.setPoints(points);
        values.clear();
        for(int i = 0; i < points.length;++i)
            values.add(null);
    }
    
    
    /**
     * Adds a group of points and their associated values.
     * @param points The points to add as an array of 3 or 4-element arrays.
     * @param values The values to use as an array of equal length to points.
     * @throws sddm.SDDMException If the two parameters are not of the same length.
     */
    public void add(float[][] points,T[] values) throws SDDMException {
        if (values.length != points.length)
            throw new SDDMException("Number of points does not match number of values while attempting to add points to PointMap");
        try {
            super.addPoints(points);
            for(int i = 0; i < values.length;++i)
                this.values.add(values[i]);
        } catch (SDDMException e) {
            //keep number of values consistent
            int toAdd = this.list.size() - this.values.size();
            for(int i = 0; i < toAdd;++i)
                this.values.add(values[i]);
            throw e;
        }
    }
    
    /**
     * Adds a group of points with a value of null.
     * @param points The points to add as an array of 3 or 4-element arrays.
     */
    public void addPoints(float[][] points) {
        try {
            super.addPoints(points);
            for(int i = 0; i < points.length;++i)
                values.add(null);
        } catch (SDDMException e) {
            //keep number of values consistent
            int toAdd = this.list.size() - this.values.size();
            for(int i = 0; i < toAdd;++i)
                this.values.add(null);
        }
    }
    
    /**
     * Adds a point with its associated value.
     * @param point The point to add as an {x,y,z} or {x,y,z,w} array.
     * @param value The value to assign to add associated with the point.
     * @throws sddm.SDDMException If point is not 3 or 4 elements long.
     */
    public void add(float[] point,T value) throws SDDMException {
        super.addPoint(point);
        values.add(value);
    }
    
    /**
     * Adds a point with a null value.
     * @param point The point to add as an {x,y,z} or {x,y,z,w} array.
     * @throws sddm.SDDMException If point is not 3 or 4 elements long.
     */
    public void addPoint(float[] point) throws SDDMException {
        super.addPoint(point);
        values.add(null);
    }
    
    /**
     * Adds a point at the specified location with an associated value.  Uses a w-coordinate of 0.
     * @param x the x-coordinate of the point to be added.
     * @param y the y-coordinate of the point to be added.
     * @param z the z-coordinate of the point to be added.
     * @param value The value to associate with the point.
     */
    public void add(float x, float y, float z, T value) {
        super.addPoint(x, y, z);
        values.add(value);
    }
    
    /**
     * Adds a point at the specified location with a null value.
     * @param x the x-coordinate of the point to be added.
     * @param y the y-coordinate of the point to be added.
     * @param z the z-coordinate of the point to be added.
     */
    public void addPoint(float x, float y, float z) {
        super.addPoint(x, y, z);
        values.add(null);
    }
    
    /**
     * Removes the point and its value at the specified index, returning the value.
     * @param index the index at which to remove a point/value pair.
     * @return the value at the removed point.
     * @throws java.lang.IndexOutOfBoundsException If the index addresses an element beyond the length of this PointMap.
     */
    public T remove(int index) throws IndexOutOfBoundsException {
        super.removePoint(index);
        return values.remove(index);
    }
    
    /**
     *   Removes the first point in the list with the specified set of coordinates, returning the value.
     *
     * @return the value at the removed point, or null if no such point is present.
     * @param point a 3 or 4-element array as {x,y,z} or {x,y,z,w}.
     * @throws sddm.SDDMException if the supplied point is not 3 or 4 elements long.
     */
    public T remove(float[] point)  throws SDDMException {
        int index = this.indexOf(point);
        if (index == -1)
            return null;
        this.list.remove(index);
        return this.values.remove(index);
    }
    
    /**
     * Removes the first point in the list with the specified set of x,y, and z coordinates, returning the value.
     * @param x the x-coordinate of the point to be removed.
     * @param y the y-coordinate of the point to be removed.
     * @param z the z-coordinate of the point to be removed.
     * @return the value at the removed point, or null if no such point is present.
     */
    public T remove(float x, float y, float z)  {
        int index = this.indexOf(x,y,z);
        if (index == -1)
            return null;
        this.list.remove(index);
        return this.values.remove(index);
    }
    
    /**
     * Removes the first point in the list with the specified set of x,y, z, and w coordinates, returning the value.
     * @param x the x-coordinate of the point to be removed.
     * @param y the y-coordinate of the point to be removed.
     * @param z the z-coordinate of the point to be removed.
     * @param w the w-coordinate of the point to be removed.
     * @return the value at the point removed, or null if no such point is present.
     */
    public T remove(float x, float y, float z,float w)  {
        int index = this.indexOf(x,y,z,w);
        if (index == -1)
            return null;
        this.list.remove(index);
        return this.values.remove(index);
    }
    
    /**
     * Removes a point from this PointMap (and its value) and returns a reference to the array storing the point.
     * @param index the index of the point to remove
     * @throws java.lang.IndexOutOfBoundsException If the index addresses an element beyond the length of this PointMap.
     * @return the location of the point as a 4-element array
     */
    public float[] removePoint(int index) throws IndexOutOfBoundsException  {
        values.remove(index);
        return super.removePoint(index);
    }
    
    /**
     *
     * Retrieves the value at the specified index
     * @param index the index from which to retrieve the value.
     * @return the value from this PointMap
     * @throws java.lang.IndexOutOfBoundsException If the index addresses an element beyond the length of this PointMap.
     */
    public T get(int index) throws IndexOutOfBoundsException {
        return values.get(index);
    }
    
    /**
     * Retrieves the value at the first point with the specified coordinates.
     * @param point A 3 or 4-element array to compare coordinates against.
     * @return the first value at a location matching the supplied point, or null if there is no such point.
     */
    public T get(float[] point) {
        int index = this.indexOf(point);
        return this.values.get(index);
    }
    
    /**
     * Retrieves the value at the first point with the specified coordinates.
     * @param x the x-coordinate of the point to search for.
     * @param y the y-coordinate of the point to search for.
     * @param z the z-coordinate of the point to search for.
     * @return the first value at a location matching the supplied point, or null if there is no such point.
     */
    public T get(float x, float y, float z) {
        int index = this.indexOf(x,y,z);
        return this.values.get(index);
    }
    
    /**
     *
     * Retrieves the value at the first point with the specified coordinates.
     * @param x the x-coordinate of the point to search for.
     * @param y the y-coordinate of the point to search for.
     * @param z the z-coordinate of the point to search for.
     * @param w the w-coordinate of the point to search for.
     * @return the first value at a location matching the supplied point, or null if there is no such point.
     */
    public T get(float x, float y, float z,float w) {
        int index = this.indexOf(x,y,z,w);
        return this.values.get(index);
    }
    
    /**
     * Generates a copy of the value in this PointMap at the specified index.
     * @param index the index from which to copy the value
     * @return a copy of the value
     * @throws sddm.SDDMException If the value type doesn't support a public clone() method or there is some other problem cloning the element.
     * @throws java.lang.IndexOutOfBoundsException If the index addresses an element beyond the length of this PointMap.
     */
    public T copy(int index)  throws SDDMException, IndexOutOfBoundsException {
        T result = values.get(index);
        java.lang.reflect.Method cloneMethod;
        try{
            cloneMethod = result.getClass().getMethod("clone");
        } catch (Exception e) {
            cloneMethod = null;
        }
        
        if (cloneMethod != null) {
            try {
                return (T) cloneMethod.invoke(result);
            } catch (Exception e) {
                throw new SDDMException("Problem trying to copy an element of a Point Map that doesn't have a clone method:\n"+e,e);
            }
        } else
            throw new SDDMException("Tried to copy an element of a Point Map that doesn't have a clone method:\n"+result.toString());
    }
    
    /**
     *
     * Generates a copy of the point in this PointMap at the specified index.
     * @param index The index from which to get the point.
     * @return a copy of the point as a 4-element array.
     * @throws java.lang.IndexOutOfBoundsException If the index addresses an element beyond the length of this PointMap.
     */
    public float[] copyPoint(int index)  throws IndexOutOfBoundsException {
        return super.copyPoint(index);
    }
    
    /**
     *
     * Clones this PointMap, deeply copying the points, and deeply copying the references if a clone method is available, or shallowly if not.
     * @return a copy of this PointMap
     */
    public PointMap clone() {
        PointMap newMap = new PointMap();
        //copy PointList and related statistics
        PointList pl = super.clone();
        newMap.list = pl.list;
        newMap.stats = pl.stats;
        newMap.compor = this.compor;
        
        //Copy values deeply if cloneable, or shallowly if there's  a problem in the cloning process
        for(T elem : this.values)  {
            //TODO: clean this up for things that aren't cloneable
            T toAdd;
            try {
                java.lang.reflect.Method m = elem.getClass().getMethod("clone");
                toAdd = (T)m.invoke(elem);
            } catch (Exception e) {
                toAdd = elem;
            }
            
            newMap.values.add(toAdd);
        }
        return newMap;
    }
    
    /**
     * Removes all elements from this PointMap.
     */
    public void clear() {
        super.clear();
        values.clear();
    }
    
    /**
     * Adds a point at the specified location with an associated value.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param z the z-coordinate of the point.
     * @param w the w-coordinate of the point.
     * @param value The value to associate with the point.
     */
    public void add(float x, float y, float z, float w, T value) {
        super.addPoint(x, y, z, w);
        values.add(value);
    }
    
    /**
     * Adds the specified point with a null value.
     * @param x the x-coordinate of the point to add.
     * @param y the y-coordinate of the point to add.
     * @param z the z-coordinate of the point to add.
     * @param w the w-coordinate of the point to add.
     */
    public void addPoint(float x, float y, float z, float w) {
        super.addPoint(x, y, z, w);
        values.add(null);
    }
    
    /**
     * Generates a list of the points and values in this PointMap
     * @return a string with all the points and their associated value on their own line
     */
    public String toString() {
        StringBuffer sb = new StringBuffer("Point List with Elements:\n");
        
        for ( int i = 0; i < this.list.size(); ++i) {
            float[] f = this.list.get(i);
            sb.append(f[0]);
            sb.append(", ");
            sb.append(f[1]);
            sb.append(", ");
            sb.append(f[2]);
            sb.append(", ");
            sb.append(f[3]);
            sb.append(", value:");
            sb.append(this.values.get(i));
            sb.append('\n');
        }
        
        return sb.toString();
    }
    
    /**
     * generates a hash code from the sum of the backing objects.
     * @return a hash code
     */
    public int hashCode() {
        return super.hashCode()+values.hashCode();
    }
    
    /**
     *
     * Generates an iterator that iterates over all the values in this PointMap
     * @return a new iterator with remove capability
     */
    public Iterator<T> valueIterator() {
        return new Iterator<T>() {
            int index = -1;
            boolean removed = false;
            public void remove() {
                if (index == -1 )
                    throw new IllegalStateException("Called remove before next.");
                if (removed)
                    throw new IllegalStateException("Called remove twice after a call to next");
                values.remove(index);
                list.remove(index);
                removed = true;
            }
            
            public T next() {
                removed = false;
                return values.get(++index);
            }
            
            public boolean hasNext() {
                return (index + 1 < values.size());
            }
            
        };
    }
    
    java.lang.reflect.Method compareMethod = null;
    java.lang.reflect.Method floatMethod  = null;
    
    /**
     *
     * Determines the largest value from the compareTo methods of the PointMap's values.
     * @return the maximum value in this PointMap
     * @throws sddm.SDDMException if there is a problem invoking the compareTo method
     * @throws java.lang.NoSuchMethodException If there is no compareTo method
     */
    public T getMaxValue() throws SDDMException, NoSuchMethodException {
        return this.values.get(getIndexOfMaxValue(compor));
    }
    
    /**
     *
     * Determines the smallest value from the compareTo methods of the PointMap's values.
     * @return the minimum value in this PointMap
     * @throws sddm.SDDMException if there is a problem invoking the compareTo method
     * @throws java.lang.NoSuchMethodException If there is no compareTo method
     */
    public T getMinValue() throws SDDMException, NoSuchMethodException {
        
        return this.values.get(getIndexOfMinValue(compor));
    }
    
    /**
     * Determines the average based on a float value determined by a call to either float floatVal(),float toFloat(), float float(), or float length()
     *
     * @return the average of the values in this PointMap (or 0 if it is empty)
     * @throws java.lang.NoSuchMethodException if none of the of the appropriate methods (see above) are present
     * @throws sddm.SDDMException if there is a problem determining the value (i.e. if nont of the methods are present in T)
     */
    public float getAvgValue() throws SDDMException, NoSuchMethodException {
        if (floatMethod == null && !this.list.isEmpty())
            floatMethod = this.findFloatMethod(this.list.get(0).getClass());
        
        float avg = 0;
        try {
            for(T elem: this.values)
                avg+=((Float)floatMethod.invoke(elem));
        } catch (Exception e) {
            throw new SDDMException("Problem determining average value:"+e.getMessage(),e);
        }
        
        avg = avg / this.values.size();
        
        return avg;
    }
    
    /**
     * Uses the int CompareTo(Object o) method from {@link Comparable} to determine the maximum value in this PointMap
     *
     * @return the index of the maximum value in this PointMap (as determined by the CompareTo method) or -1 if it is empty
     * @throws java.lang.NoSuchMethodException If there is no compareTo method
     * @throws sddm.SDDMException if there is a problem invoking the compareTo method
     */
    public int getIndexOfMaxValue() throws SDDMException, NoSuchMethodException {
        return getIndexOfMaxValue(null);
    }
    
    /**
     * Uses a supplied {@link Comparator} to determine the maximum value in this PointMap
     *
     * @param comp The comparator to use, or null if the objects are {@link Comparable} (i.e. uses their compareTo method)
     * @throws sddm.SDDMException if there is a problem invoking the compareTo method (for null comp)
     * @throws java.lang.NoSuchMethodException If there is no compareTo method (for null comp)
     * @return the index of the maximum value in this PointMap or -1 if it is empty
     */
    public int getIndexOfMaxValue(Comparator comp) throws SDDMException, NoSuchMethodException {
        int index = 0;
        
        if (comp == null) {
            if (compareMethod == null )
                compareMethod = this.findCompareToMethod(this.values.get(0).getClass());
            
            
            try {
                T currMax = this.values.get(0);
                //TODO:test to make sure I don't have this backwards
                for (int i = 1; i < this.values.size(); ++i) {
                    T currVal = this.values.get(i);
                    if( (Integer)compareMethod.invoke(currMax,currVal) <  0 ) {
                        currMax = currVal;
                        index = i;
                    }
                }
            } catch(Exception e) {
                if (this.values.isEmpty())
                    return -1;
                throw new SDDMException("Problem searching for max value:"+e.getMessage(),e);
            }
        } else {
            //TODO:implement comp
            try {
                T currMax = this.values.get(0);
                //TODO:test to make sure I don't have this backwards
                for (int i = 1; i < this.values.size(); ++i) {
                    T currVal = this.values.get(i);
                    if( comp.compare(currMax,currVal) <  0 ) {
                        currMax = currVal;
                        index = i;
                    }
                }
            } catch(Exception e) {
                if (this.values.isEmpty())
                    return -1;
                throw new SDDMException("Problem searching for max value with comparator:"+e.getMessage(),e);
            }
        }
        return index;
    }
    
    /**
     * Uses the int CompareTo(Object o) method from {@link Comparable} to determine the minimum value in this PointMap
     *
     * @return the index of the minimum value in this PointMap (as determined by the CompareTo method) or -1 if it is empty
     * @throws java.lang.NoSuchMethodException If there is no compareTo method
     * @throws sddm.SDDMException if there is a problem invoking the compareTo method
     */
    public int getIndexOfMinValue() throws SDDMException, NoSuchMethodException {
        return getIndexOfMinValue(null);
    }
    
    /**
     * Uses a supplied {@link Comparator} to determine the minimum value in this PointMap
     *
     * @param comp The comparator to use, or null if the objects are {@link Comparable} (i.e. uses their compareTo method)
     * @throws sddm.SDDMException if there is a problem invoking the compareTo method (for null comp)
     * @throws java.lang.NoSuchMethodException If there is no compareTo method (for null comp)
     * @return the index of the minimum value in this PointMap or -1 if it is empty
     */
    public int getIndexOfMinValue(Comparator comp) throws SDDMException, NoSuchMethodException {
        int index = 0;
        
        if (comp == null) {
            if (compareMethod == null )
                compareMethod = this.findCompareToMethod(this.values.get(0).getClass());
            
            try {
                T currMin = this.values.get(0);
                //TODO:test to make sure I don't have this backwards
                for (int i = 1; i < this.values.size(); ++i) {
                    T currVal = this.values.get(i);
                    if( (Integer)compareMethod.invoke(currMin,currVal) >  0 ) {
                        currMin = currVal;
                        index = i;
                    }
                }
            } catch(Exception e) {
                if (this.values.isEmpty())
                    return -1;
                throw new SDDMException("Problem searching for min value:"+e.getMessage(),e);
            }
        } else {
            //TODO:implement
            try {
                T currMin = this.values.get(0);
                //TODO:test to make sure I don't have this backwards
                for (int i = 1; i < this.values.size(); ++i) {
                    T currVal = this.values.get(i);
                    if( comp.compare(currMin,currVal) >  0 ) {
                        currMin = currVal;
                        index = i;
                    }
                }
            } catch(Exception e) {
                if (this.values.isEmpty())
                    return -1;
                throw new SDDMException("Problem searching for min value with comparator:"+e.getMessage(),e);
            }
        }
        return index;
    }

    /**
     * Generates a PointMap from a file with a list of 3D or 4D points and a single float value (i.e. x0 y0 z0 val0 x1 y1 z1 val1 x2 y2 z2 val2 ... or x0 y0 z0 w0 val0 x1 y1 z1 w1 val1 x2 y2 z2 w2 val2 ...).
     * @param f the file to load the values from
     * @param threeElement if true, takes the points in the file to be represented by three numbers (the fourth is set to 1.0). If false, takes 4 values for the point's coordinates before the value for that point.
     * @param field if the format is a field (i.e. rectilinear grid).  If so, the expected file format has size of each of the dimensions as the first three elements, and the rest should be the values in x-major,y-semi-major,z-minor order.???
     * @throws sddm.FileFormatException if the number of elements in the file does not match the required number, or some other problem loading the data in the file.
     * @throws java.io.FileNotFoundException if the specified file does not exist
     * @return the PointMap representation of the supplied file
     */
    public static PointMap<Float> loadPointsAndFloatsFromFile(File f, boolean threeElement, boolean field) throws FileFormatException, FileNotFoundException {
        PointMap<Float> pm = new PointMap<Float>();
        
        java.util.Scanner s = new java.util.Scanner(f);
        
        if (field) {
            int xSize = 0;
            int ySize = 0;
            int zSize = 0;
            int wSize = 0;
            try {
                xSize = (int)s.nextFloat();
                ySize = (int)s.nextFloat();
                zSize = (int)s.nextFloat();
                wSize = 1;
                if (!threeElement)
                    wSize = (int)s.nextFloat();
            } catch (Exception e) {
                throw new FileFormatException("Could not Load dimensions of field; first three elements must be xsize,ysize,zsize.",e);
            }
            int i = 0;
            int j = 0;
            int k = 0;
            int l = 1; // w-coordinate begins at 1 due to divide by zero issues in homogenous coordinates
            while(s.hasNextFloat()) {
                float val = s.nextFloat();
                pm.add(i,j,k,l,val);
                if (++i >= xSize) {
                    i=0;
                    if (++j >= ySize) {
                        j=0;
                        if (++k >= zSize) {
                            k=0;
                            if (++l > wSize + 1)
                                throw new FileFormatException("Too many elements in file to account for expected dimension("+xSize+"X"+ySize+"X"+zSize+"X"+wSize+")");
                        }
                    }
                }
            }
            if (l != wSize + 1)
                throw new FileFormatException("Not enough elements in file to account for expected dimension("+xSize+"X"+ySize+"X"+zSize+"X"+wSize+")");
            
        } else {
            while(s.hasNextFloat()) {
                try {
                    float x = s.nextFloat();
                    float y = s.nextFloat();
                    float z = s.nextFloat();
                    float w = 1.0f;
                    if (threeElement)
                        w=s.nextFloat();
                    float value = s.nextFloat();
                    
                    pm.add(x,y,z,w,value);
                    
                } catch (NoSuchElementException e) {
                    throw new FileFormatException("Number of numbers in file " +f+ " not divisible by "+ (threeElement ? "4":"5") );
                } catch (Exception e) {
                    throw new FileFormatException("Problem loading file:"+e.getMessage(),e);
                }
            }
        }
        return pm;
    }
    /**
     * Generates a PointMap from a file with a list of three-diemnsional points and a single float value (i.e. x0 y0 z0 val0 x1 y1 z1 val1 x2 y2 z2 val2 ...).
     * @param f the file to load the values from
     * @throws sddm.FileFormatException if the number of elements in the file does not match the required number, or some other problem loading the data in the file.
     * @throws java.io.FileNotFoundException if the specified file does not exist
     * @return the PointMap representation of the supplied file
     */
    public static PointMap<Float> loadPointsAndFloatsFromFile(File f) throws FileFormatException, FileNotFoundException {
        return PointMap.loadPointsAndFloatsFromFile(f,true,false);
    }
    
    /**
     * Generates a PointMap from a file with a list of 3 or 4D points and a 3 or 4D vector (i.e. x0 y0 z0 a0 b0 c0 x1 y1 z1 a1 b1 c1 x2 y2 z2 a2 b2 c2 ... or x0 y0 z0 w0 a0 b0 c0 d0 x1 y1 z1 w1 a1 b1 c1 d1 x2 y2 z2 w2 a2 b2 c2 d2... ).
     * @param f the file to load the values from
     * @param threeElement if true, takes the points and vectors in the file to be represented by three numbers (the fourth is set to 1.0). If false, takes 4 values for the point's coordinates and 4 coordinated for the vector at that point.
     * @param field if the format is a field (i.e. rectilinear grid).  If so, the expected file format has size of each of the dimensions as the first three elements, and the rest should be the values in x-major,y-semi-major,z-minor order.???
     * @throws sddm.FileFormatException if the number of elements in the file does not match the required number, or some other problem loading the data in the file.
     * @throws java.io.FileNotFoundException if the specified file does not exist
     * @return the PointMap representation of the supplied file
     */
    public static PointMap<javax.vecmath.Vector4f> loadPointsAndFloatVectorsFromFile(File f, boolean threeElement, boolean field) throws FileFormatException, FileNotFoundException {
        PointMap<javax.vecmath.Vector4f> pm = new PointMap<javax.vecmath.Vector4f>();
        
        java.util.Scanner s = new java.util.Scanner(f);
        if (field) {
            int xSize = 0;
            int ySize = 0;
            int zSize = 0;
            int wSize = 0;
            try {
                xSize = (int)s.nextFloat();
                ySize = (int)s.nextFloat();
                zSize = (int)s.nextFloat();
                wSize = 1;
                if (!threeElement)
                    wSize = (int)s.nextFloat();
            } catch (Exception e) {
                throw new FileFormatException("Could not Load dimensions of field; first three elements must be xsize,ysize,zsize.",e);
            }
            int i = 0;
            int j = 0;
            int k = 0;
            int l = 1; // w-coordinate begins at 1 due to divide by zero issues in homogenous coordinates
            while(s.hasNextFloat()) {
                //TODO: add a bit more format error-checking code
                float vx = s.nextFloat();
                float vy = s.nextFloat();
                float vz = s.nextFloat();
                float vw = 1.0f;
                if (!threeElement)
                    vw = s.nextFloat();
                pm.add(i,j,k,l,new javax.vecmath.Vector4f(vx,vy,vz,vw));
                if (++i >= xSize) {
                    i=0;
                    if (++j >= ySize) {
                        j=0;
                        if (++k >= zSize) {
                            k=0;
                            if (++l > wSize + 1)
                                throw new FileFormatException("Too many elements in file to account for expected dimension("+xSize+"X"+ySize+"X"+zSize+"X"+wSize+")");
                        }
                    }
                }
            }
            if (l != wSize + 1)
                throw new FileFormatException("Not enough elements in file to account for expected dimension("+xSize+"X"+ySize+"X"+zSize+"X"+wSize+")");
            
        } else {
            while(s.hasNextFloat()) {
                try {
                    float x = s.nextFloat();
                    float y = s.nextFloat();
                    float z = s.nextFloat();
                    float w = 1.0f;
                    if (threeElement)
                        w = s.nextFloat();
                    
                    float vx = s.nextFloat();
                    float vy = s.nextFloat();
                    float vz = s.nextFloat();
                    float vw = 1.0f;
                    if (threeElement)
                        vw = s.nextFloat();
                    
                    pm.add( x,y,z,w,new javax.vecmath.Vector4f(vx,vy,vz,vw) );
                    
                } catch (NoSuchElementException e) {
                    throw new FileFormatException("Number of numbers in file " +f+ " not divisible by "+ (threeElement ? "4":"5") );
                } catch (Exception e) {
                    throw new FileFormatException("Problem loading file:"+e.getMessage(),e);
                }
            }
        }
        pm.setComparator(new Comparator<javax.vecmath.Vector4f>() {
            public int compare(javax.vecmath.Vector4f v1, javax.vecmath.Vector4f v2) {
                return Float.compare(v1.length(),v2.length());
            }
            
        });
        return pm;
    }
    
    /**
     * Generates a PointMap from a file with a list of 3-dimensional points and a 3-dimensional vector (i.e. x0 y0 z0 a0 b0 c0 x1 y1 z1 a1 b1 c1 x2 y2 z2 a2 b2 c2 ... ).
     * @param f the file to load the values from
     * @throws sddm.FileFormatException if the number of elements in the file does not match the required number, or some other problem loading the data in the file.
     * @throws java.io.FileNotFoundException if the specified file does not exist
     * @return the PointMap representation of the supplied file
     */
    public static PointMap<javax.vecmath.Vector4f> loadPointsAndFloatVectorsFromFile(File f) throws FileFormatException, FileNotFoundException {
        return PointMap.loadPointsAndFloatVectorsFromFile(f,true,false);
    }
    
    //-------------privates-------------------
    
    private static java.lang.reflect.Method findCompareToMethod(Class c) throws NoSuchMethodException {
        return c.getMethod("compareTo",Object.class);
    }
    
    private static java.lang.reflect.Method findFloatMethod(Class c) throws NoSuchMethodException {
        //TODO:implement
        //throw SDDMException if it doesn't exist
        //use float floatValue(),float toFloat(), float float(), or float length()
        java.lang.reflect.Method m;
        
        try {
            m = c.getMethod("floatValue");
        } catch(NoSuchMethodException e1) {
            try {
                m = c.getMethod("toFloat");
            } catch(NoSuchMethodException e2) {
                try {
                    m = c.getMethod("float");
                } catch (NoSuchMethodException e3) {
                    m = c.getMethod("length");
                    //if none are present, pass the exception back through to the caller method
                }
            }
        }
        
        return m;
    }
    
}
